/*
 * Copyright (C) 1997-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "server.h"
#include "config.h"
#include <sys/mman.h>
/* #include <time.h> */

#define BACKLOG 10 /*  how many pending connections queue will hold */

struct mp3_meta_data *mmd = NULL;
static FILE *logfile = NULL;

struct gengetopt_args_info conf;
static int mmd_semid;
static int signal_pipe[2];

char *afs_buf = NULL;

/*
 * setup shared memory area and get semaphore for locking
 */
static void init_shm(void)
{
	int fd;
	caddr_t area;

	if ((fd = open("/dev/zero", O_RDWR)) < 0) {
		clog(EMERG, "%s: failed to open /dev/zero\n", __func__);
		exit(EXIT_FAILURE);
	}
	if ((area = mmap(0, sizeof(struct mp3_meta_data),
			PROT_READ | PROT_WRITE,
			MAP_SHARED, fd, 0)) == (caddr_t) - 1) {
		clog(EMERG, "%s: mmap error\n", __func__);
		exit(EXIT_FAILURE);
	}
	close(fd); /* we dont need /dev/zero anymore */
	mmd = (struct mp3_meta_data *)area;
	
	mmd->num_played = 0;
	mmd->num_commands = 0;
	mmd->active_connections = 0;
	mmd->stream = -1;
	mmd->old_stream = -1;
	mmd->afs_status_flags = AFS_NEXT;
	mmd->lcl_head = NULL;

        mmd_semid = semget(IPC_PRIVATE, 1, IPC_CREAT | 0666);
        if (mmd_semid == -1) {
                clog(EMERG, "%s: semget failed\n", __func__);
		exit(EXIT_FAILURE);
        }
}

void do_semop(struct sembuf *sops, int num)
{
        if (semop(mmd_semid, sops, num) >= 0)
		return;
	clog(WARNING, "%s: semop failed (%s), retrying\n", __func__,
		strerror(errno));
	while (semop(mmd_semid, sops, num) < 0)
		; /* nothing */
}

/*
 * lock mmd
 */
void mmd_lock(void)
{
        struct sembuf sops[2] = {
                {
			.sem_num = 0,
			.sem_op = 0,
			.sem_flg = SEM_UNDO
		},
                {
			.sem_num = 0,
			.sem_op = 1,
			.sem_flg = SEM_UNDO
		}
        };
	
	do_semop(sops, 2);
}

/*
 * unlock mmd
 */
void mmd_unlock(void)
{
        struct sembuf sops[1] = {
                {
			.sem_num = 0,
			.sem_op = -1,
			.sem_flg = SEM_UNDO
		},
        };
        do_semop(sops, 1);
}

void mmd_dup(struct mp3_meta_data *new_mmd)
{
	mmd_lock();
	*new_mmd = *mmd;
	mmd_unlock();
}

/*
 * construct string containing uptime
 */
void uptime_str(int i, char *msg)
{
	int n;

	snprintf(msg, MAXLINE - 1, "%i day%s, %i:%02i:%02i%n", i / 86400,
		i > 86400 && i < 2 * 86400?  "" : "s",
		(i / 3600) % 24, (i / 60) % 60, i % 60, &n);
	msg[n] = '\0';
}

/*
 * keep track of # connections the server had by now.
 */
unsigned int num_connects(int get_or_incr)
{
	static unsigned int num_connections;

	if (get_or_incr != INCR)
		return num_connections;
	else
		return ++num_connections;
}

/*
 * set or get server's uptime
 */
int server_uptime(int set_or_get)
{
	static time_t startuptime;
	time_t now;

	if (set_or_get == SET) {
		time(&startuptime);
		return 0;
	}
	time(&now);
	return((int) difftime(now, startuptime));
}

/* Open user_list file, returns pointer to opened file on success,
 * NULL on errors
 */
static FILE *open_user_list(char *config_file)
{
	FILE *cf;

	clog(DEBUG, "%s: Opening config file %s\n", __func__, config_file);
	if (!(cf = fopen(config_file, "r"))) {
		clog(INFO, "%s: Failed to open user list %s (%s)\n", __func__,
			config_file, strerror(errno));
	        return NULL;
	}
	return cf;
}
/*
 * lookup user in user_list file. Fills in a user struct containing
 * filename of the user's public key as well as the permissions of that user.
 * Returns 1 on success, 0 if user does not exist and -1 on errors.
 */
static int get_user(struct user *user) {
	FILE *file_ptr;
	char *char_ptr;
	char line[MAXLINE];
	/* keyword, user, key, perms */
	char w[MAXLINE], n[MAXLINE], k[MAXLINE], p[MAXLINE], tmp[4][MAXLINE];
	int num;

	file_ptr = open_user_list(mmd->user_list);
	if (!file_ptr)
		return -1;
	while (fgets(line, MAXLINE, file_ptr)) {
		clog(PETER_B, "%s: Read line (%i bytes) "
			"from config file\n", __func__, strlen(line));
		if (sscanf(line,"%s %s %s %s", w, n, k, p) < 3)
			continue;
		if (!strcmp(w, "user") && !strcmp(user->name, n)) {
			clog(DEBUG, "%s: found entry for %s\n", __func__, n);
			strcpy(user->name, n);
			strcpy(user->pubkey_file, k);
			user->perms = 0;
			char_ptr = p;
			num=sscanf(char_ptr, "%[A-Z_],%[A-Z_],%[A-Z_],%[A-Z_]",
				tmp[0], tmp[1], tmp[2], tmp[3]);
			clog(DEBUG, "%s: found %i perm entries\n", __func__,
				num);
			user->perms = 0;
			while (num > 0) {
				num--;
				clog(DEBUG, "%s: tmp[%i]=%s\n", __func__,
					num, tmp[num]);
				if (!strcmp(tmp[num], "AFS_READ"))
					user->perms =
						user->perms | AFS_READ;
				else if (!strcmp(tmp[num], "AFS_WRITE"))
					user->perms =
						user->perms | AFS_WRITE;
				else if (!strcmp(tmp[num], "DB_READ"))
					user->perms = user->perms | DB_READ;
				else if (!strcmp(tmp[num], "DB_WRITE"))
					user->perms = user->perms | DB_WRITE;
				else /* unknown permission */
					clog(WARNING, "%s: unknown permission:"
						"%s\n", __func__, tmp[num]);
			}
			fclose(file_ptr);
			return 1;
		}
	}
	fclose(file_ptr);
	return 0;
}

/*
 * encrypts challgenge number with public rsa key of user. encrypted
 * number is stored in buf. Returns 0 on success.
 */
static ssize_t encrypt(char *pubkey_file, long unsigned challenge_nr,
	char *buf)
{
	ssize_t n;
	int fd1[2], fd2[2];
	pid_t pid;

	buf[0] = '\0';
	if (pipe(fd1) < 0 || pipe(fd2) < 0) {
		clog(ERROR, "%s: pipe error\n", __func__);
		return -1;
	}
	if ((pid = fork()) < 0) {
		clog(ERROR, "%s: fork error\n", __func__);
		return -1;
	} else if (pid > 0) {
		/* parent */
		char *line = buf;
		ssize_t num;

		close(fd1[0]);
		close(fd2[1]);
		snprintf(line, MAXLINE - 1, "%lu\n%n", challenge_nr, &n);
		line[n] = '\0';
		clog(DEBUG, "%s: Writing challenge to pipe "
			"(%i bytes)\n", __func__, n);
		if (write(fd1[1], line, n) != n) {
			clog(ERROR, "%s: write error\n", __func__);
			return -1;
		}
		close(fd1[1]);
		clog(DEBUG, "%s: Reading line from pipe \n", __func__);
		if ((num = read(fd2[0], line, MAXLINE)) < 0) {
			clog(ERROR, "%s: read error from pipe\n", __func__);
			return -1;
		}
		line[num] = '\0';
		clog(DEBUG, "%s: Read %i byte from pipe\n", __func__, num);
	 	close(fd2[0]);
		return num;
	} else { /* child*/
		close(fd1[1]);
		close(fd2[0]);
		if (fd1[0] != STDIN_FILENO) {
			clog(DEBUG, "%s: dup2ing stdin\n", __func__);
			if (dup2(fd1[0], STDIN_FILENO) != STDIN_FILENO ) {
				clog(ERROR, "%s: dup2 error to stdin\n", __func__);
				close(fd1[0]);
			}
		}
		if (fd2[1] != STDOUT_FILENO) {
			clog(DEBUG, "%s: dup2ing stdout\n", __func__);
			if (dup2(fd2[1], STDOUT_FILENO) != STDOUT_FILENO ) {
				clog(ERROR, "%s: dup2 error to stdout\n", __func__);
				close(fd2[1]);
			}
		}
		if (execl(OPENSSL, "openssl", "rsautl", "-encrypt",
				"-pubin", "-inkey", pubkey_file, NULL) < 0) {
			clog(ERROR, "%s: execl error\n", __func__);
			exit(EXIT_FAILURE);
		}
	}
	return -1; /* never reached */
}

/*
 * do the usual stuff to become a daemon
 */
static void daemon_init(void)
{
	pid_t pid;
	
	clog(INFO, "%s: daemonizing\n", __func__);
	if ((pid = fork()) < 0) {
		clog(EMERG, "%s: Failed to init daemon\n", __func__);
		exit(EXIT_FAILURE);
	} else if (pid)
		exit(EXIT_SUCCESS); /* parent exits */
	/* child */
	setsid(); /* become session leader */
	chdir("/");
	umask(0);
}

/*
 * die on sigint/sigterm. Kill all children too.
 */
void sigint_sigterm_handler(int s)
{
	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM); /*  nec. to get rid of the children */
	exit(s);
}

/*
 * just write one integer to signal pipe
 */
static void generic_signal_handler(int s)
{
	write(signal_pipe[1], &s, sizeof(int));
}

static void handle_sigchld(void)
{
	int status;
	pid_t pid;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		if (WIFEXITED(status)) {
			clog(DEBUG, "%s: child %i exited. Exit status: %i\n",
				__func__, pid, WEXITSTATUS(status));
			continue;
		}
		if (WIFSIGNALED(status)) {
			clog(DEBUG, "%s: child %i was killed by signal %i\n",
				__func__, pid, WTERMSIG(status));
			continue;
		}
		clog(WARNING, "%s: child %i terminated abormally\n",
			__func__, pid);
	}
}

/*
 * log function. first argument is loglevel.
 */
void clog(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";
	pid_t mypid;

	if (mmd && ll < conf.loglevel_arg)
		return;
	if (!logfile) {
		if (ll < ERROR)
			outfd = stdout;
		else
			outfd = stderr;
	} else
		outfd = logfile;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);	
	if (conf.loglevel_arg <= VERBOSE)
		fprintf(outfd, "%i: ", ll);
	mypid = getpid();
	if (conf.loglevel_arg <= VERBOSE)
		fprintf(outfd, "(%d) ", mypid);
	if (conf.loglevel_arg > VERBOSE) {
		fmt = strchr(fmt, ':');
		if (!fmt)
			return;
		fmt++;
		fmt++;
	}
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
	return;
}

/*
 * nuke stream list, freeing used memory
 */
static void stream_list_nuke(void)
{
	char **ptr = mmd->stream_list;

	clog(INFO, "%s: nuking stream list\n", __func__);
	while (*ptr) {
		clog(DEBUG, "%s: nuking %s\n", __func__, *ptr);
		free(*ptr);
		ptr++;
	}
	free(mmd->stream_list);
}

/*
 * initialise stream list.
 */
void stream_list_init(void)
{
	char **stream_list_ptr;
	FILE *pipe = NULL;
        char buf[MAXLINE];
        char stream[MAXLINE];
	int num_streams = 0, i;
	char *streams_cmd[] = {"streams", NULL};

	if (!(pipe = popen_read_dbtool(streams_cmd)))
		exit(EXIT_FAILURE);
        while (fgets(buf, MAXLINE - 1, pipe) && buf[0] != '#')
		num_streams++;
	pclose(pipe);
	num_streams++;
	if (!(pipe = popen_read_dbtool(streams_cmd)))
		exit(EXIT_FAILURE);
	clog(DEBUG, "%s: mallocing mem for %i streamnames\n", __func__,
		num_streams);
	stream_list_ptr = malloc(num_streams * sizeof(char *));
	if (!stream_list_ptr)
		exit(EXIT_FAILURE);
	clog(DEBUG, "%s: reading streams\n", __func__);
	i = 0;
        while (i < num_streams  && fgets(buf, MAXLINE - 1, pipe)) {

		buf[MAXLINE - 1] = '\0';
		if (sscanf(buf, "%200s", stream) < 1)
				continue;
		stream_list_ptr[i] = strdup(stream);
		if (!stream_list_ptr[i]) {
			clog(EMERG, "%s: strdup error\n", __func__);
			exit(EXIT_FAILURE);
		}
		clog(DEBUG, "%s: added stream %s\n", __func__,
			stream_list_ptr[i]);
		i++;
	}
	pclose(pipe);
	if (!i)
		clog(CRIT, "%s: no streams found\n", __func__);
	stream_list_ptr[i] = NULL;
	mmd->stream_list = stream_list_ptr;
}

void open_log(void)
{
	if (!(logfile = fopen(mmd->logfile, "a"))) {
		clog(EMERG, "%s: opening logfile %s failed (%s)", __func__,
			mmd->logfile, strerror(errno));
		exit(EXIT_FAILURE);
	}
	setlinebuf(logfile);
	clog(INFO, "%s: logfile %s opened\n", __func__, mmd->logfile);
}
void close_log(void)
{
	if (logfile) {
		clog(INFO, "%s: closing logfile\n", __func__);
		fclose(logfile);
		logfile = NULL;
	}
}

void reload_log(void)
{
	close_log();
	open_log();
}

void parse_config(int override)
{
	char *home;
	struct gengetopt_args_info *args_info = &conf;
 //       static char default_config_file[_POSIX_PATH_MAX];
   //     static char default_user_list[_POSIX_PATH_MAX];
        struct stat statbuf;
        int ret;

        if (!args_info->config_file_given || !args_info->user_list_given) {
                if (!(home = getenv("HOME"))) {
                        clog(EMERG, "%s: getenv error, cannot read HOME\n",
				__func__);
                        exit(EXIT_FAILURE);
                }
	}
        if (!args_info->config_file_given)
                sprintf(mmd->config_file, "%s/.paraslash/server.conf", home);
        else {
		snprintf(mmd->config_file, sizeof(mmd->config_file) - 1,
			"%s%n", mmd->config_file, &ret);
		mmd->config_file[ret] = '\0';
	}
        ret = stat(mmd->config_file, &statbuf);
        if (ret && args_info->config_file_given) {
                fprintf(stderr, "Can't stat config file %s\n",
                        mmd->config_file);
                exit(EXIT_FAILURE);
        }
        if (!ret)
                cmdline_parser_configfile(mmd->config_file, args_info,
			override, 0);
	/* user list */
        if (!args_info->user_list_given)
		sprintf(mmd->user_list, "%s/.paraslash/server.users", home);
	else {
		snprintf(mmd->user_list, sizeof(mmd->user_list) - 1,
			"%s%n", args_info->user_list_arg, &ret);
		mmd->user_list[ret] = '\0';
	}
	/* logfile */
	if (args_info->logfile_given) {
		snprintf(mmd->logfile, sizeof(mmd->logfile) - 1,
			"%s%n", args_info->logfile_arg, &ret);
		mmd->logfile[ret] = '\0';
		open_log();
	} else {
		if (conf.daemon_given) {
			clog(EMERG, "%s: daemon mode, but no logfile\n",
				__func__);
			exit(EXIT_FAILURE);
		}
		logfile = NULL;
	}
	/* afs _buf size */
	if (args_info->afs_buf_given)
		if (args_info->afs_buf_arg < 0) {
			clog(EMERG, "%s: invalid buffer size\n", __func__);
			exit(EXIT_FAILURE);
		}
}
void alloc_afs_buf(void)
{

	if (afs_buf)
		free(afs_buf);
	afs_buf = malloc(conf.afs_buf_arg * sizeof(char));
	if (!afs_buf) {
		clog(EMERG, "%s: afs_buf malloc error\n", __func__);
		exit(EXIT_FAILURE);
	}
}

/*
 * called when server gets SIGHUP. Reread configuration, then close and
 * reopen log. Finally, send SIGHUP to all processes found in the pid list
 * to force them to reread log too.
 */
static int handle_sighup(void)
{
	char *stream = NULL, *old_stream = NULL;

	clog(NOTICE, "%s: Got SIGHUP\n", __func__);
	close_log();
	parse_config(1);

	/* reread linked list of commands and stream list */
	lcl_nuke();
	lcl_init();
	
	/* reread list of admissible streams */
	if (mmd->stream >= 0) { /* save old values*/
		stream = strdup(mmd->stream_list[mmd->stream]);
		old_stream = strdup(mmd->stream_list[mmd->old_stream]);
	}
	stream_list_nuke();
	stream_list_init();
	if (stream) {
		mmd->stream = find_stream(stream, mmd);
		mmd->old_stream = find_stream(old_stream, mmd);
		free(stream);
		free(old_stream);
	}
	/* re-allocate afs_buf, it's size might have changed */
	alloc_afs_buf();
	return 0;
}

static void install_sighandler(int sig, void (*fct)(int), int restart)
{
	struct sigaction sa;

	clog(DEBUG, "%s: Installing handler for signal %d\n", __func__, sig);
	sa.sa_handler = fct;
	sigemptyset(&(sa.sa_mask));
	if (restart)
		sa.sa_flags = SA_RESTART;
	if (sigaction(sig, &sa, NULL) == -1) {
		clog(EMERG, "%s: Failed to install sighandler %d\n", __func__,
			sig);
		exit(EXIT_FAILURE);
	}
}

static void install_sighandlers(void)
{
	if (pipe(signal_pipe) < 0) {
		clog(ERROR, "%s: %s", __func__, strerror(errno));
		exit(EXIT_FAILURE);
        }
	install_sighandler(SIGINT, &sigint_sigterm_handler, 0);
	install_sighandler(SIGTERM, &sigint_sigterm_handler, 0);
	install_sighandler(SIGHUP, &generic_signal_handler, 0);
	install_sighandler(SIGCHLD, &generic_signal_handler, 1);
	signal(SIGPIPE, SIG_IGN);
}

/*
 * read random input from /dev/urandom into n
 */
static int read_urandom(long unsigned *n)
{	
	int fd, ret = 0;

	if (!(fd = open("/dev/urandom", O_RDONLY)))
		return 0;
	if (read(fd, n, sizeof(long unsigned)))
		ret = 1;
	close(fd);
	return ret;
}


static void handle_connect(int new_fd)
{
	int numbytes, ret, argc;
	char buf[STRINGSIZE];
	struct user user_struct;
	char command[MAXLINE] = "", user[MAXLINE] = "";
	struct linked_cmd_list *cmd_ptr;
	long unsigned challenge_nr, chall_response;
	char **argv;

	if (!read_urandom(&challenge_nr)) {
		clog(ERROR, "%s: urandom error\n", __func__);
		goto err_out;
	}
	/* send Welcome message */
	sprintf(buf, "This is para_server, version %s.\n", VERSION);
	if (send_buffer(new_fd, buf) < 0)
		goto err_out;
	/* recv auth request line */
	if ((numbytes = recv_buffer(new_fd, buf, sizeof(buf) - 1) <= 0))
		goto err_out;
	if (sscanf(buf, "%8s %32s", command, user) < 2 ||
		strncmp(command, "auth", 4) || !user[0]) {
		clog(WARNING, "%s: auth error\n", __func__);
		send_buffer(new_fd, "Protocol mismatch.\n");
		goto err_out;
	}
	clog(DEBUG, "%s: Received auth request for user %s\n", __func__,
		user);
	strncpy(user_struct.name, user, sizeof(user_struct.name));
	user_struct.name[sizeof(user_struct.name) - 1] = '\0';

	/* lookup user in list file */
	if ((ret = get_user(&user_struct)) < 0) {
		clog(DEBUG, "%s: Error reading user list\n", __func__);
		goto err_out;
	}
	if (!ret) { /* user not found */
		clog(WARNING, "%s: Auth request for unknown user %s\n",
			__func__, user);
		sprintf(buf, "You don't exist. Go away!\n");
		send_buffer(new_fd, buf);
		goto err_out;
	}
	/* encrypt challenge_nr with public key */
	if ((numbytes = (encrypt(user_struct.pubkey_file, challenge_nr,
			buf))) <= 0)
		goto err_out;
	clog(DEBUG, "%s: Sending %d byte challenge for %s\n", __func__,
		numbytes, user);
	/* We can't use send_buffer here since buf may contain null bytes */
	if (send_bin_buffer(new_fd, buf, numbytes) < 0)
		goto err_out;
	/* recv decrypted number */
	if ((numbytes = recv_buffer(new_fd, buf, sizeof(buf) - 1)) <= 0)
		goto err_out;
	if (sscanf(buf, "challenge_response:%lu", &chall_response) < 1
			|| chall_response != challenge_nr) {
		send_buffer(new_fd, "Authentication failed\n");
		clog(WARNING, "%s: bad auth, aborting (got %s)\n", __func__, buf);
		goto err_out;
	}
	/* auth successful. Send 'Proceed' message */
	clog(INFO, "%s: Good auth for %s (%lu)\n", __func__, user,
		challenge_nr);
	if (send_buffer(new_fd, PROCEED_MSG "\n") < 0)
		goto err_out;
	/* read command */
	if ((numbytes = recv_buffer(new_fd, buf, sizeof(buf) - 1)) <= 0)
		goto err_out;
	/* parse command */
	if (!(cmd_ptr = parse_cmd(buf))) {
		send_buffer(new_fd, "Unknown command\n");
		goto err_out;
	}
	/* valid command, check permissions */
	ret = check_perms(user_struct.perms, cmd_ptr);
	if (ret < 0) {
		/* insufficient perms */
		clog(WARNING, "%s: insufficient permissions "
			"for command %s\n", __func__, cmd_ptr->cmd->name);
		send_buffer(new_fd, "Permission denied\n");
		goto err_out;
	}
	/* valid command and sufficient perms */
	argc = split_args(buf, &argv, '\n');
	if (!argv) {
		clog(ERROR, "%s: can't split arguments\n", __func__);
		goto err_out;
	}
	argv[0] = cmd_ptr->cmd->name;
	mmd_lock();
	mmd->num_commands++;
	if (cmd_ptr->cmd->perms & DB_WRITE)
		mmd->afs_status_flags |= AFS_REREAD_DB_INFO;
	mmd_unlock();
	clog(NOTICE, "%s: executing command handler for command \"%s\"\n",
		__func__, cmd_ptr->cmd->name);
	ret = cmd_ptr->cmd->handler(new_fd, argc, argv) < 0? EXIT_FAILURE :
		EXIT_SUCCESS;
	goto out;
err_out:
	ret = EXIT_FAILURE;
out:
	mmd_lock();
	mmd->active_connections--;
	mmd_unlock();
	exit(ret);
}

static int init_network(void)
{
	int sockfd;
	struct sockaddr_in my_addr;

	clog(INFO, "%s: init network socket\n", __func__);
	if ((sockfd = get_socket()) < 0)
		exit(EXIT_FAILURE);
	if (setserversockopts(sockfd) < 0)
		exit(EXIT_FAILURE);
	init_sockaddr(&my_addr, conf.port_arg, NULL);
	if (do_bind(sockfd, &my_addr) < 0)
		exit(EXIT_FAILURE);
	if (listen(sockfd, BACKLOG) == -1) {
		clog(EMERG, "%s: listen failed\n", __func__);
		exit(EXIT_FAILURE);
	}
	return sockfd;
}


int do_inits(int argc, char **argv)
{
	/* connector's address information */
	int sockfd;
	
	/* init mmd struct */
	init_shm();
	/* parse command line options */
	cmdline_parser(argc, argv, &conf);
	/* parse config file, open log and set defaults */
	parse_config(0);
	/* buffer for the audio file sender */
	alloc_afs_buf();
	/* reset server uptime */
	server_uptime(SET);
	/* print/log welcome message */
	if (conf.loglevel_arg < NOTICE)
		cmdline_parser_print_version();
	clog(NOTICE, "%s: ---- para_server " VERSION " startup ----\nbuild: %s",
		__func__, BUILD_DATE);
	/* become daemon */
	if (conf.daemon_given)
		daemon_init();
	mmd->server_pid = getpid();
	/* setup signal handlers */
	install_sighandlers();
	/* init linked list of commands */
	lcl_init();
	/* init list of all admissible streams */
	stream_list_init();
	/* default stream */
	if (conf.sstream_given) {
		mmd->stream = find_stream(conf.sstream_arg, mmd);
		mmd->old_stream = mmd->stream;
	} else if (*mmd->stream_list) {
		mmd->stream = 0;
		mmd->old_stream = 0;
	}
	/* init network socket */
	sockfd = init_network();
	mmd_lock();
	return sockfd;
}

/*
 * taken from libdaemon
 */
int daemon_signal_next(void)
{
	int s;
	ssize_t r;

	if ((r = read(signal_pipe[0], &s, sizeof(s))) == sizeof(s))
		return s;
	if (r < 0) {
		if (errno == EAGAIN)
			return 0;
		else {
			clog(ERROR, "%s: %s", __func__, strerror(errno));
			return -1;
		}
	}
	clog(ERROR, "%s: Short read() on signal pipe.", __func__);
	return -1;
}

/*
 * MAIN
 */
int main(int argc, char **argv)
{
	/* listen on sock_fd, new connection on new_fd */
	int sockfd, new_fd;
	FILE *poc_pipe = NULL;
	struct sockaddr_in their_addr;

	sockfd = do_inits(argc, argv);
	/* main accept() loop */
	while(1) {
		int i, poc_fileno, max_fileno, ret, check_poc = 0;
		pid_t chld_pid;
		fd_set rfds, wfds;

		/* check socket for reading */
		FD_ZERO(&rfds);
		FD_SET(sockfd, &rfds);
		max_fileno = sockfd;

		/* check signal pipe for reading */
		FD_SET(signal_pipe[0], &rfds);
		max_fileno = MAX(max_fileno, signal_pipe[0]);

		check_poc = 0;
		if (mmd->afs_status_flags & AFS_PLAYING) {
			afs_preselect();
			/* do we still want to play? */
			if (mmd->afs_status_flags & AFS_PLAYING) {
				/* check poc pipe for writing */
				check_poc = 1;
				poc_pipe = get_stream_writer_pipe(conf.stream_write_cmd_arg);
				poc_fileno = fileno(poc_pipe);
				FD_ZERO(&wfds);
				FD_SET(poc_fileno, &wfds);
				max_fileno = MAX(poc_fileno, max_fileno);
			}
		}
		mmd_unlock();
		ret = select(max_fileno + 1, &rfds, check_poc? &wfds : NULL,
			NULL, NULL);
		mmd_lock();
		if (mmd->afs_status_flags & AFS_REREAD_DB_INFO)
			update_dbinfo();

		if (ret <= 0)
			continue;
		if (FD_ISSET(signal_pipe[0], &rfds)) {
			int sig_nr;
			sig_nr = daemon_signal_next();
			clog(DEBUG, "%s: Caught signal %d\n", __func__, sig_nr);
			switch(sig_nr) {
			case SIGHUP:
				handle_sighup();
				break;
			case SIGCHLD:
				handle_sigchld();
				break;
			}
		}

		if (check_poc && FD_ISSET(poc_fileno, &wfds))
			afs_send_chunk(poc_fileno);
		if (!FD_ISSET(sockfd, &rfds))
			continue;
		
		new_fd = do_accept(sockfd, &their_addr);
		if (new_fd < 0)
			continue;
		clog(NOTICE, "%s: Got connection from %s\n", __func__,
			inet_ntoa(their_addr.sin_addr));
		num_connects(INCR);
		mmd->active_connections++;
		chld_pid = fork();
		if (chld_pid < 0) {
			clog(ERROR, "%s: fork failed\n", __func__);
			continue;
		}
		if (chld_pid) {
			close(new_fd);
			/* parent keeps accepting connections */
			continue;
		}
		/* child doesn't need the listener and ignores SIGCHLD */
		close(sockfd);
		signal(SIGCHLD, SIG_IGN);
		/*
		 * put info on who we are serving into argv[0] to make
		 * process visible for top/ps
		 */
		for (i = argc - 1; i >= 0; i--)
			memset(argv[i], 0, strlen(argv[i]));
		sprintf(argv[0], "para_server (serving %s)",
			inet_ntoa(their_addr.sin_addr));
		handle_connect(new_fd);
	}
	return 0;
}
